from typing import Awaitable, Callable, Dict, Any
import websockets
from arkitekt.agents.transport.base import AgentTransport
import asyncio
import json
from arkitekt.agents.transport.errors import (
    AgentTransportException,
    AssignationListDeniedError,
    ProvisionListDeniedError,
)
from arkitekt.agents.transport.protocols.agent_json import *
import logging
from websockets.exceptions import ConnectionClosedError, InvalidStatusCode
from arkitekt.api.schema import LogLevelInput

from koil.types import ContextBool, Contextual

logger = logging.getLogger(__name__)


async def token_loader():
    raise NotImplementedError(
        "Websocket transport does need a defined token_loader on Connection"
    )


class CorrectableConnectionFail(AgentTransportException):
    pass


class DefiniteConnectionFail(AgentTransportException):
    pass


class WebsocketAgentTransport(AgentTransport):
    ws_url: str
    instance_id: Optional[str]
    token_loader: Callable[[], Awaitable[str]] = Field(exclude=True)
    max_retries = 5
    time_between_retries = 5
    allow_reconnect = True

    _futures: Contextual[Dict[str, asyncio.Future]] = None
    _connected: ContextBool = False
    _healthy: ContextBool = False
    _send_queue: Contextual[asyncio.Queue] = None
    _connection_task: Contextual[asyncio.Task] = None

    async def __aenter__(self):
        assert self._abroadcast is not None, "Broadcast ss be defined"
        self._futures = {}
        self._send_queue = asyncio.Queue()
        self._connection_task = asyncio.create_task(self.websocket_loop())
        self._connected = True

    async def websocket_loop(self, retry=0, reload_token=False):
        send_task = None
        receive_task = None
        try:
            try:
                token = await self.token_loader(force_refresh=reload_token)

                async with websockets.connect(
                    f"{self.ws_url}?token={token}&instance_id={self.instance_id}"
                ) as client:

                    logger.info("Agent on Websockets connected")

                    send_task = asyncio.create_task(self.sending(client))
                    receive_task = asyncio.create_task(self.receiving(client))

                    self._healthy = True
                    done, pending = await asyncio.wait(
                        [send_task, receive_task],
                        return_when=asyncio.FIRST_EXCEPTION,
                    )
                    self._healthy = False

                    for task in pending:
                        task.cancel()

                    for task in done:
                        raise task.exception()

            except InvalidStatusCode as e:
                logger.warning(
                    "Websocket Connect was denied. Trying to reload token",
                    exc_info=True,
                )
                reload_token = True
                raise CorrectableConnectionFail from e

            except ConnectionClosedError as e:
                logger.warning("Websocket was closed", exc_info=True)
                raise CorrectableConnectionFail from e

            except Exception as e:
                logger.warning("Websocket excepted. Trying to recover", exc_info=True)
                raise CorrectableConnectionFail from e

        except CorrectableConnectionFail as e:
            logger.info(f"Trying to Recover from Exception {e}")
            if retry > self.max_retries or not self.allow_reconnect:
                raise DefiniteConnectionFail("Exceeded Number of Retries")

            await asyncio.sleep(self.time_between_retries)
            logger.info(f"Retrying to connect")
            await self.websocket_loop(retry=retry + 1, reload_token=reload_token)

        except DefiniteConnectionFail as e:
            logger.error("Websocket excepted closed definetely", exc_info=True)
            self.connection_dead = False
            raise e

        except asyncio.CancelledError as e:
            logger.info("Websocket got cancelled. Trying to shutdown graceully")
            if send_task and receive_task:
                send_task.cancel()
                receive_task.cancel()

            cancellation = await asyncio.gather(
                send_task, receive_task, return_exceptions=True
            )
            raise e

    async def sending(self, client):
        try:
            while True:
                message = await self._send_queue.get()
                await client.send(message)
                self._send_queue.task_done()
        except asyncio.CancelledError as e:
            logger.info("Sending Task sucessfully Cancelled")

    async def receiving(self, client):
        try:
            async for message in client:
                logger.info(f"Receiving message {message}")
                await self.receive(message)
        except asyncio.CancelledError as e:
            logger.info("Receiving Task sucessfully Cancelled")

    async def receive(self, message):
        json_dict = json.loads(message)
        if "type" in json_dict:
            type = json_dict["type"]
            id = json_dict["id"]

            # State Layer
            if type == AgentSubMessageTypes.ASSIGN:
                await self._abroadcast(AssignSubMessage(**json_dict))

            if type == AgentSubMessageTypes.UNASSIGN:
                await self._abroadcast(UnassignSubMessage(**json_dict))

            if type == AgentSubMessageTypes.UNPROVIDE:
                await self._abroadcast(UnprovideSubMessage(**json_dict))

            if type == AgentSubMessageTypes.PROVIDE:
                await self._abroadcast(ProvideSubMessage(**json_dict))

            if type == AgentMessageTypes.LIST_ASSIGNATIONS_REPLY:
                self._futures[id].set_result(AssignationsListReply(**json_dict))
            if type == AgentMessageTypes.LIST_ASSIGNATIONS_DENIED:
                self._futures[id].set_exception(
                    AssignationListDeniedError(json_dict["error"])
                )

            if type == AgentMessageTypes.LIST_PROVISIONS_REPLY:
                self._futures[id].set_result(ProvisionListReply(**json_dict))
            if type == AgentMessageTypes.LIST_PROVISIONS_DENIED:
                self._futures[id].set_exception(
                    ProvisionListDeniedError(json_dict["error"])
                )

        else:
            logger.error(f"Unexpected messsage: {json_dict}")

    async def awaitaction(self, action: JSONMessage):
        assert self._connected, "Websocket is not connected"
        if action.id in self._futures:
            raise ValueError("Action already has a future")

        future = asyncio.Future()
        self._futures[action.id] = future
        await self._send_queue.put(action.json())
        return await future

    async def delayaction(self, action: JSONMessage):
        assert self._connected, "Websocket is not connected"
        await self._send_queue.put(action.json())

    async def list_provisions(
        self, exclude: Optional[ProvisionStatus] = None
    ) -> List[Provision]:
        action = ProvisionList(exclude=exclude)
        prov_list_reply: ProvisionListReply = await self.awaitaction(action)
        return prov_list_reply.provisions

    async def change_provision(
        self,
        id: str,
        status: ProvisionStatus = None,
        message: str = None,
        mode: ProvisionMode = None,
    ):
        action = ProvisionChangedMessage(
            provision=id, status=status, message=message, mode=mode
        )
        await self.delayaction(action)

    async def change_assignation(
        self,
        id: str,
        status: AssignationStatus = None,
        message: str = None,
        returns: List[Any] = None,
    ):
        action = AssignationChangedMessage(
            assignation=id, status=status, message=message, returns=returns
        )
        await self.delayaction(action)

    async def list_assignations(
        self, exclude: Optional[AssignationStatus] = None
    ) -> List[Assignation]:
        action = AssignationsList(exclude=exclude)
        ass_list_reply: AssignationsListReply = await self.awaitaction(action)
        return ass_list_reply.assignations

    async def log_to_assignation(
        self, id: str, level: LogLevelInput = None, message: str = None
    ):
        action = AssignationLogMessage(assignation=id, level=level, message=message)
        await self.delayaction(action)

    async def log_to_provision(
        self, id: str, level: LogLevelInput = None, message: str = None
    ):
        action = ProvisionLogMessage(provision=id, level=level, message=message)
        await self.delayaction(action)

    async def __aexit__(self, *args, **kwargs):
        self._connection_task.cancel()
        self._connected = False

        try:
            await self._connection_task
        except asyncio.CancelledError:
            pass



from arkitekt.postmans.transport.base import PostmanTransport
from arkitekt.messages import (
    Assignation,
    Reservation,
    Unassignation,
    Unreservation,
)
from arkitekt.api.schema import (
    AssignationStatus,
    ReservationStatus,
    ReserveParamsInput,
)
from typing import Any, Dict, List, Optional, Union
import asyncio
import random
from pydantic import Field
from koil import unkoil


class MockAutoresolvingPostmanTransport(PostmanTransport):
    """A mock transport for an agent

    Args:
        AgentTransport (_type_): _description_
    """

    assignationState: Dict[str, Assignation] = Field(default_factory=dict)
    unassignationState: Dict[str, Unassignation] = Field(default_factory=dict)
    unreservationState: Dict[str, Unreservation] = Field(default_factory=dict)
    reservationState: Dict[str, Reservation] = Field(default_factory=dict)

    _task: Optional[asyncio.Task] = None

    async def alist_assignations(
        self, exclude: Optional[AssignationStatus] = None
    ) -> List[Assignation]:
        return self.assignationState.values()

    async def alist_reservations(
        self, exclude: Optional[ReservationStatus] = None
    ) -> List[Reservation]:
        return self.reservationState.values()

    async def __aenter__(self):
        self._task = asyncio.create_task(self.aresolve_reservations())

    async def aassign(
        self,
        reservation: str,
        args: List[Any],
        kwargs: Dict[str, Any],
        persist=True,
        log=False,
    ) -> Assignation:

        assignation = Assignation(
            assignation=str(len(self.assignationState) + 1),
            reservation=reservation,
            args=args,
            kwargs=kwargs,
            status=AssignationStatus.PENDING,
        )
        self.assignationState[assignation.assignation] = assignation
        return assignation

    async def areserve(
        self, node: str, params: ReserveParamsInput = None
    ) -> Reservation:
        reservation = Reservation(
            reservation=str(len(self.reservationState) + 1),
            node=node,
            status=ReservationStatus.ROUTING,
        )
        self.reservationState[reservation.reservation] = reservation
        return reservation

    async def aunreserve(self, reservation: str) -> Unreservation:
        self.reservationState[reservation].update(
            Reservation(reservation=reservation, status=ReservationStatus.CANCELLED)
        )
        return Unreservation(reservation=reservation)

    async def aunassign(self, assignation: str) -> Unassignation:
        self.assignationState[assignation].update(
            Assignation(assignation=assignation, status=AssignationStatus.CANCELLED)
        )
        return assignation

    async def aresolve_reservations(self):

        while True:

            await asyncio.sleep(0.1)

            ress = [
                res
                for key, res in self.reservationState.items()
                if res.status == ReservationStatus.ROUTING
            ]
            if ress:
                res = random.choice(ress)
                res.status = ReservationStatus.ACTIVE

                self.reservationState[res.reservation] = res
                await self._abroadcast(res)

            asss = [
                ass
                for key, ass in self.assignationState.items()
                if ass.status == AssignationStatus.PENDING
            ]

            if asss:
                ass = random.choice(asss)
                ass.status = AssignationStatus.RETURNED
                ass.returns = []

                self.assignationState[ass.assignation] = ass
                await self._abroadcast(ass)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._task.cancel()

        try:
            await self._task
        except asyncio.CancelledError as e:
            pass

    class Config:
        underscore_attrs_are_private = True


class MockPostmanTransport(PostmanTransport):

    assignationState: Dict[str, Assignation] = Field(default_factory=dict)
    unassignationState: Dict[str, Unassignation] = Field(default_factory=dict)
    unreservationState: Dict[str, Unreservation] = Field(default_factory=dict)
    reservationState: Dict[str, Reservation] = Field(default_factory=dict)

    _inqueue: asyncio.Queue = None

    async def alist_assignations(
        self, exclude: Optional[AssignationStatus] = None
    ) -> List[Assignation]:
        return []

    async def alist_reservations(
        self, exclude: Optional[ReservationStatus] = None
    ) -> List[Reservation]:
        return []

    async def __aenter__(self):
        self._inqueue = asyncio.Queue()

    async def aassign(
        self,
        reservation: str,
        args: List[Any],
        kwargs: Dict[str, Any],
        persist=True,
        log=False,
    ) -> Assignation:

        assignation = Assignation(
            assignation=str(len(self.assignationState) + 1),
            reservation=reservation,
            args=args,
            kwargs=kwargs,
            status=AssignationStatus.PENDING,
        )
        self.assignationState[assignation.assignation] = assignation
        await self._inqueue.put(assignation)
        return assignation

    async def areserve(
        self, node: str, params: ReserveParamsInput = None
    ) -> Reservation:
        print("Called")
        reservation = Reservation(
            reservation=str(len(self.reservationState) + 1),
            node=node,
            status=ReservationStatus.ROUTING,
        )
        self.reservationState[reservation.reservation] = reservation
        await self._inqueue.put(reservation)
        return reservation

    async def aunreserve(self, reservation: str) -> Unreservation:
        unreservation = Unreservation(reservation=reservation)
        await self._inqueue.put(unreservation)
        return unreservation

    async def aunassign(self, assignation: str) -> Unassignation:
        unassignation = Unassignation(assignation=assignation)
        await self._inqueue.put(Unassignation(assignation=assignation))
        return unassignation

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        for item in range(self._inqueue.qsize()):
            self._inqueue.task_done()

    async def adelay(
        self, message: Union[Assignation, Reservation, Unreservation, Unassignation]
    ):
        await self._abroadcast(message)

    async def areceive(self, timeout=None):
        if timeout:
            return await asyncio.wait_for(self._inqueue.get(), timeout)
        return await self._inqueue.get()

    def delay(
        self, message: Union[Assignation, Reservation, Unreservation, Unassignation]
    ):
        return unkoil(self.adelay, message)

    def receive(self, *args, **kwargs):
        return unkoil(self.areceive, *args, **kwargs)

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True



from typing import Awaitable, Callable, Type, TypeVar

T = TypeVar("T")


async def idshrinker(self):
    return self.id


def wraps(
    identifier: str,
    expander: Callable[[Type[T]], Awaitable[T]],
    shrinker: Callable[[Type[T]], Awaitable[T]] = idshrinker,
) -> Callable[[Type], Type]:
    def real_cls_decorator(cls: Type[T]) -> Type[T]:
        assert isinstance(cls, Type), "This is a class decorator"

        cls.get_identifier = identifier
        cls.ashrink = shrinker
        cls.aexpand = expander

        return cls

    return real_cls_decorator



from graphql import OperationType
from rath.links import compose
from rath.links.aiohttp import AIOHttpLink
from rath.links.auth import AuthTokenLink
from rath.links.dictinglink import DictingLink
from rath.links.shrink import ShrinkingLink
from arkitekt.rath import ArkitektRath
from herre import Herre, current_herre
from fakts import Config
from rath.links.split import SplitLink
from rath.links.websockets import WebSocketLink


class ArkitektRathConfig(Config):
    endpoint_url: str
    ws_endpoint_url: str

    class Config:
        group = "arkitekt"


class FaktsArkitektRath(ArkitektRath):
    def configure(self, config: ArkitektRathConfig, herre: Herre) -> None:
        self.link = compose(
            ShrinkingLink(),
            DictingLink(),
            AuthTokenLink(token_loader=herre.aget_token),
            SplitLink(
                left=AIOHttpLink(url=config.endpoint_url),
                right=WebSocketLink(
                    url=config.ws_endpoint_url, token_loader=herre.aget_token
                ),
                split=lambda o: o.node.operation != OperationType.SUBSCRIPTION,
            ),
        )

    async def __aenter__(self):

        herre = current_herre.get()

        config = await ArkitektRathConfig.from_fakts()
        self.configure(config, herre)

        return await super().__aenter__()



from arkitekt.structures.registry import StructureRegistry
from arkitekt.api.schema import TemplateFragment, NodeFragment, aget_template, afind

DEFAULT_STRUCTURE_REGISTRY = None


def get_default_structure_registry() -> StructureRegistry:
    global DEFAULT_STRUCTURE_REGISTRY
    if not DEFAULT_STRUCTURE_REGISTRY:
        DEFAULT_STRUCTURE_REGISTRY = StructureRegistry()

        DEFAULT_STRUCTURE_REGISTRY.register_as_structure(
            TemplateFragment, "@arkitekt/template", expand=aget_template
        )
        DEFAULT_STRUCTURE_REGISTRY.register_as_structure(
            NodeFragment, "@arkitekt/node", expand=afind
        )

    return DEFAULT_STRUCTURE_REGISTRY



from arkitekt.agents.stateful import StatefulAgent
from arkitekt.agents.transport.websocket import WebsocketAgentTransport
from arkitekt.rath import ArkitektRath
from fakts import Config
from herre.herre import Herre, current_herre
from arkitekt.rath import ArkitektRath, current_arkitekt_rath


class AgentConfig(Config):
    endpoint_url: str
    instance_id: str

    class Config:
        group = "arkitekt.agent"


class FaktsAgent(StatefulAgent):
    def configure(self, config: AgentConfig, herre: Herre, rath: ArkitektRath) -> None:

        self.transport = WebsocketAgentTransport(
            ws_url=config.endpoint_url,
            instance_id=config.instance_id,
            token_loader=herre.aget_token,
        )

        self.rath = rath

    async def __aenter__(self):

        config = await AgentConfig.from_fakts()
        self.configure(
            config, herre=current_herre.get(), rath=current_arkitekt_rath.get()
        )

        return await super().__aenter__()



from qtpy import QtWidgets, QtGui, QtCore
from arkitekt.app import ArkitektApp
from koil.qt import QtRunner
from koil.composition.qt import QtPedanticKoil
from .utils import get_image_path


class MagicBar(QtWidgets.QWidget):
    def __init__(self, app: ArkitektApp, dark_mode: bool = False) -> None:
        super().__init__()
        self.app = app
        assert isinstance(self.app.koil, QtPedanticKoil), "Koil should be Qt Koil"
        self.dark_mode = dark_mode

        self.connect_task = QtRunner(self.app.aconnect)
        self.connect_task.errored.connect(print)
        self.connect_task.returned.connect(self.on_connected)

        self.provide_task: QtRunner = QtRunner(self.app.arkitekt.arun)
        self.provide_task.errored.connect(print)
        self.provide_task.returned.connect(self.on_providing_ended)

        self.magicb = QtWidgets.QPushButton("Connect")
        self.magicb.setMinimumHeight(30)
        self.magicb.setMaximumHeight(30)

        self.connect_future = None
        self.provide_future = None

        self.layout = QtWidgets.QHBoxLayout()
        self.gearb_pix = QtGui.QPixmap(get_image_path("gear.png", dark_mode=dark_mode))
        self.gearb = QtWidgets.QPushButton()
        self.gearb.setIcon(QtGui.QIcon(self.gearb_pix))
        self.gearb.setMinimumWidth(30)
        self.gearb.setMaximumWidth(30)
        self.gearb.setMinimumHeight(30)
        self.gearb.setMaximumHeight(30)

        self.magicb.clicked.connect(self.magic_button_clicked)
        self.gearb.clicked.connect(self.gear_button_clicked)

        self.layout.addWidget(self.magicb)
        self.layout.addWidget(self.gearb)
        self.setLayout(self.layout)

        if not self.app.fakts.healthy:
            self.set_unkonfigured()
        else:
            self.set_unconnected()

    def on_connected(self):
        self.magicb.setText("Provide")

    def on_providing_ended(self):
        pass

    def gear_button_clicked(self):
        pass

    def update_movie(self):
        self.magicb.setIcon(QtGui.QIcon(self.magicb_movie.currentPixmap()))

    def set_button_movie(self, movie):
        self.magicb_movie = QtGui.QMovie(
            get_image_path(movie, dark_mode=self.dark_mode)
        )
        self.magicb_movie.frameChanged.connect(self.update_movie)
        self.magicb_movie.setScaledSize(QtCore.QSize(30, 30))
        self.magicb_movie.start()

    def set_unkonfigured(self):
        self.set_button_movie("pink pulse.gif")
        self.magicb.setDisabled(False)
        self.magicb.setText("Konfigure App")

    def set_unconnected(self):
        self.set_button_movie("green pulse.gif")
        self.magicb.setDisabled(False)
        self.magicb.setText("Connect")

    def magic_button_clicked(self):
        if not self.connect_future:
            self.connect_future = self.connect_task.run()
            self.magicb.setText("Cancel Login")
            return
        if not self.connect_future.done():
            self.connect_future.cancel()
            self.magicb.setText("Login")
            return

        if not self.provide_future:
            self.provide_future = self.provide_task.run()
            self.magicb.setText("Cancel Providing...")
            return
        if not self.provide_future.done():
            self.provide_future.cancel()
            self.magicb.setText("Provide")
            return



from rath import rath
import contextvars

current_arkitekt_rath = contextvars.ContextVar("current_arkitekt_rath", default=None)


class ArkitektRath(rath.Rath):
    """_summary_

    Args:
        rath (_type_): _description_
    """

    async def __aenter__(self):
        await super().__aenter__()
        current_arkitekt_rath.set(self)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await super().__aexit__(exc_type, exc_val, exc_tb)
        current_arkitekt_rath.set(None)



from pydantic import BaseModel
from arkitekt.actors.base import Actor
from arkitekt.api.schema import LogLevelInput
from arkitekt.messages import Assignation, Provision


class AssignationHelper(BaseModel):
    actor: Actor
    assignation: Assignation
    provision: Provision

    async def alog(self, level: LogLevelInput, message: str) -> None:
        raise NotImplementedError()


class ProvisionHelper(BaseModel):
    actor: Actor
    provision: Provision

    async def alog(self, level: LogLevelInput, message: str) -> None:
        raise NotImplementedError()


class ThreadedAssignationHelper(AssignationHelper):
    pass


class AsyncAssignationHelper(AssignationHelper):
    async def alog(self, message: str, level: LogLevelInput = LogLevelInput.DEBUG):
        await self.actor.transport.log_to_assignation(
            id=self.assignation.assignation, level=level, message=message
        )


class AsyncProvisionHelper(AssignationHelper):
    async def alog(self, message: str, level: LogLevelInput = LogLevelInput.DEBUG):
        await self.actor.transport.log_to_assignation(
            id=self.assignation.assignation, level=level, message=message
        )



from typing import Dict, Optional, Union

from pydantic import BaseModel, Field
from arkitekt.agents.transport.base import AgentTransport
from arkitekt.structures.registry import (
    StructureRegistry,
    get_current_structure_registry,
)
from arkitekt.rath import ArkitektRath
import asyncio
import logging
from arkitekt.api.schema import (
    AssignationStatus,
    ProvisionMode,
    ProvisionStatus,
    aget_template,
    TemplateFragment,
)
from arkitekt.messages import Assignation, Provision, Unassignation
from arkitekt.actors.errors import UnknownMessageError
from koil.types import Contextual

logger = logging.getLogger(__name__)


class Actor(BaseModel):
    provision: Provision
    transport: AgentTransport

    strict: bool = False
    expand_inputs: bool = True
    shrink_outputs: bool = True
    structure_registry: StructureRegistry = Field(
        default_factory=get_current_structure_registry
    )
    rath: Optional[ArkitektRath] = Field(default=None)

    template: Contextual[TemplateFragment] = None
    debug: bool = False

    runningAssignments: Dict[str, Assignation] = Field(default_factory=dict)

    _in_queue: Contextual[asyncio.Queue] = None
    _runningTasks: Dict[str, asyncio.Task] = {}
    _provision_task: asyncio.Task = None

    async def on_provide(self, provision: Provision, template: TemplateFragment):
        return None

    async def on_unprovide(self):
        return None

    async def on_assign(self, assignation: Assignation):
        raise (
            "Needs to be owerwritten in Actor Subclass. Never use this class directly"
        )

    async def apass(self, message: Union[Assignation, Unassignation]):
        assert hasattr(self, "_in_queue"), "Actor is currently not listening"
        await self._in_queue.put(message)

    async def arun(self):
        self._in_queue = asyncio.Queue()
        self._provision_task = asyncio.create_task(self.alisten())

    async def astop(self):
        self._provision_task.cancel()

        try:
            await self._provision_task
        except asyncio.CancelledError:
            print("Provision was cancelled")

    async def alisten(self):
        self.template = await aget_template(id=self.provision.template, rath=self.rath)
        try:
            self.template = await aget_template(
                id=self.provision.template, rath=self.rath
            )

            await self.transport.change_provision(
                self.provision.provision, status=ProvisionStatus.PROVIDING
            )

            prov_context = await self.on_provide(self.provision, self.template)

            await self.transport.change_provision(
                self.provision.provision,
                status=ProvisionStatus.ACTIVE,
                mode=ProvisionMode.DEBUG if self.debug else ProvisionMode.PRODUCTION,
            )
            logger.info(f"Actor for {self.provision}: Is now active")

            while True:
                message = await self._in_queue.get()
                logger.info(f"Actor for {self.provision}: Received {message}")

                if isinstance(message, Assignation):
                    task = asyncio.create_task(self.on_assign(message))
                    self.runningAssignments[message.assignation] = task

                elif isinstance(message, Unassignation):
                    if message.assignation in self.runningAssignments:
                        task = self.runningAssignments[message.assignation]
                        if not task.done():
                            print("Cancelling task")
                            task.cancel()
                        else:
                            logger.error("Task was already done")
                    else:
                        await self.transport.change_assignation(
                            status=AssignationStatus.CRITICAL,
                            message="Task was never assigned",
                        )
                else:
                    raise UnknownMessageError(f"{message}")

        except Exception as e:
            print(e)
            logger.exception("Actor failed", exc_info=True)
            await self.transport.change_provision(
                self.provision.provision,
                status=ProvisionStatus.CRITICAL,
                message=repr(e),
            )

            current_provision_context.set(None)

        except asyncio.CancelledError as e:
            print("We are getting cancelled here?")
            await self.transport.change_provision(
                self.provision.provision,
                status=ProvisionStatus.CANCELING,
                message=repr(e),
            )

            await self.on_unprovide()

            logger.info("Doing Whatever needs to be done to cancel!")

            cancel_assignations = [i.cancel() for i in self.runningAssignments.values()]

            for i in self.runningAssignments.values():
                try:
                    await i
                except asyncio.CancelledError:
                    pass

            await self.transport.change_provision(
                self.provision.provision,
                status=ProvisionStatus.CANCELLED,
                message=str(e),
            )
            raise e

    async def __aenter__(self):
        await self.arun()
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.astop()

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True



from typing import Optional

from pydantic import Field
from arkitekt.messages import Reservation
from arkitekt.postmans.vars import current_postman
from arkitekt.structures.registry import get_current_structure_registry
from koil.composition import KoiledModel
from koil.types import ContextBool
from .stateful import StatefulPostman
from arkitekt.api.schema import AssignationStatus, ReservationStatus, ReserveParamsInput
import asyncio
from arkitekt.traits.node import Reserve
from koil import unkoil
import logging
from arkitekt.structures.serialization.postman import shrink_inputs, expand_outputs


logger = logging.getLogger(__name__)


class ReservationContract(KoiledModel):
    # TODO:Assert that we can actually assign to this? validating that all of the nodes inputs are
    # registered in the structure registry?

    node: Reserve
    params: ReserveParamsInput = Field(default_factory=ReserveParamsInput)
    auto_unreserve: bool = False
    shrink_inputs: bool = True
    expand_outputs: bool = True

    active: ContextBool = False
    postman: Optional[StatefulPostman] = None
    _reservation: Reservation = None
    _enter_future: asyncio.Future = None
    _exit_future: asyncio.Future = None
    _updates_queue: asyncio.Queue = None
    _updates_watcher: asyncio.Task = None

    async def aassign(self, *args, structure_registry=None, **kwargs):
        assert self._reservation, "We never entered the context manager"
        assert (
            self._reservation.status == ReservationStatus.ACTIVE
        ), "Reservation is not active"

        structure_registry = structure_registry or get_current_structure_registry()
        _ass_queue = asyncio.Queue()

        shrinked_args, shrinked_kwargs = (
            await shrink_inputs(
                self.node, args, kwargs, structure_registry=structure_registry
            )
            if self.shrink_inputs
            else (args, kwargs)
        )

        ass = await self.postman.aassign(
            self._reservation.reservation, shrinked_args, shrinked_kwargs
        )
        self.postman.register_assignation_queue(ass.assignation, _ass_queue)
        logger.info(f"Listening to assignation updates for {ass}")
        try:
            while True:  # Waiting for assignation
                ass = await _ass_queue.get()
                logger.info(f"Reservation Context: {ass}")
                if ass.status == AssignationStatus.RETURNED:
                    self.postman.unregister_assignation_queue(ass.assignation)
                    outputs = (
                        await expand_outputs(
                            self.node,
                            ass.returns,
                            structure_registry=structure_registry,
                        )
                        if self.expand_outputs
                        else ass.returns
                    )
                    return outputs

                if ass.status == AssignationStatus.CRITICAL:
                    self.postman.unregister_assignation_queue(ass.assignation)
                    raise Exception(f"Critical error: {ass.message}")
        except asyncio.CancelledError as e:
            await self.postman.aunassign(ass.assignation)

            ass = await asyncio.wait_for(_ass_queue.get(), timeout=2)
            if ass.status == AssignationStatus.CANCELLED:
                logger.info("Wonderfully cancelled that assignation!")
                raise e

            raise Exception(f"Critical error: {ass}")

    def assign(self, *args, **kwargs):
        return unkoil(self.aassign, *args, **kwargs)

    async def watch_updates(self):
        try:
            if self._reservation.status == ReservationStatus.ACTIVE:
                self._enter_future.set_result(True)

            while True:
                self._reservation = await self._updates_queue.get()
                if self._reservation.status == ReservationStatus.ACTIVE:
                    if self._enter_future and not self._enter_future.done():
                        self._enter_future.set_result(True)

        except asyncio.CancelledError:
            pass

    async def __aenter__(self):
        self.postman = self.postman or current_postman.get()
        logger.info(f"Trying to reserve {self.node}")
        self._reservation = await self.postman.areserve(self.node.id, self.params)
        logger.info(f"Waiting for Reservation {self._reservation}")
        self._updates_queue = asyncio.Queue()
        self._enter_future = asyncio.Future()

        self.postman.register_reservation_queue(
            self._reservation.reservation, self._updates_queue
        )
        self._updates_watcher = asyncio.create_task(self.watch_updates())
        await self._enter_future  # Waiting to enter
        self.active = True
        return self

    async def __aexit__(self, *args, **kwargs):
        self.active = False

        if self.auto_unreserve:

            unreservation = await asyncio.wait_for(
                self.postman.aunreserve(self._reservation.reservation), timeout=1
            )
            logger.info(f"Unreserved {unreservation}")

        self._updates_watcher.cancel()

        try:
            await self._updates_watcher
        except asyncio.CancelledError:
            pass

        self.postman.unregister_reservation_queue(self._reservation.reservation)

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True


class use(ReservationContract):
    pass



from typing import Any, Dict, List, Optional
import uuid
from typing_extensions import Literal


from pydantic import BaseModel, Field
from datetime import datetime
from arkitekt.messages import Assignation, Reservation, Unassignation, Unreservation
from arkitekt.api.schema import ReservationStatus, AssignationStatus, ReserveParamsInput
from enum import Enum


class PostmanMessageTypes(str, Enum):

    LIST_RESERVATION = "RESERVE_LIST"
    LIST_RESERVATION_REPLY = "RESERVE_LIST_REPLY"
    LIST_RESERVATION_DENIED = "RESERVE_LIST_DENIED"

    RESERVE = "RESERVE"
    RESERVE_REPLY = "RESERVE_REPLY"
    RESERVE_DENIED = "RESERVE_DENIED"

    UNRESERVE = "UNRESERVE"
    UNRESERVE_REPLY = "UNRESERVE_REPLY"
    UNRESERVE_DENIED = "UNRESERVE_DENIED"

    LIST_ASSIGNATION = "ASSIGN_LIST"
    LIST_ASSIGNATION_REPLY = "ASSIGN_LIST_REPLY"
    LIST_ASSIGNATION_DENIED = "ASSIGN_LIST_DENIED"

    ASSIGN = "ASSIGN"
    ASSIGN_REPLY = "ASSIGN_REPLY"
    ASSIGN_DENIED = "ASSIGN_DENIED"

    UNASSIGN = "UNASSIGN"
    UNASSIGN_REPLY = "UNASSIGN_REPLY"
    UNASSIGN_DENIED = "UNASSIGN_DENIED"


class PostmanSubMessageTypes(str, Enum):

    ASSIGN_UPDATE = "ASSIGN_UPDATE"
    RESERVE_UPDATE = "RESERVE_UPDATE"


class JSONMeta(BaseModel):
    timestamp: datetime = Field(default_factory=datetime.utcnow)


class JSONMessage(BaseModel):
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    type: str
    meta: JSONMeta = Field(default_factory=JSONMeta)


class ReserveList(JSONMessage):
    type: Literal[
        PostmanMessageTypes.LIST_RESERVATION
    ] = PostmanMessageTypes.LIST_RESERVATION
    exclude: Optional[List[ReservationStatus]]


class ReserveListReply(JSONMessage):
    type: Literal[
        PostmanMessageTypes.LIST_RESERVATION_REPLY
    ] = PostmanMessageTypes.LIST_RESERVATION_REPLY
    reservations: List[Reservation]


class ReserveListDenied(JSONMessage):
    type: Literal[
        PostmanMessageTypes.LIST_RESERVATION_DENIED
    ] = PostmanMessageTypes.LIST_RESERVATION_DENIED
    error: str


class ReservePub(JSONMessage):
    type: Literal[PostmanMessageTypes.RESERVE] = PostmanMessageTypes.RESERVE
    params: Optional[ReserveParamsInput]
    node: Optional[str]
    template: Optional[str]
    title: Optional[str]


class ReservePubReply(JSONMessage, Reservation):
    type: Literal[PostmanMessageTypes.RESERVE_REPLY] = PostmanMessageTypes.RESERVE_REPLY


class ReservePubDenied(JSONMessage):
    type: Literal[
        PostmanMessageTypes.RESERVE_DENIED
    ] = PostmanMessageTypes.RESERVE_DENIED
    error: str


class UnreservePub(JSONMessage):
    type: Literal[PostmanMessageTypes.UNRESERVE] = PostmanMessageTypes.UNRESERVE
    reservation: str


class UnreservePubReply(JSONMessage, Unreservation):
    type: Literal[
        PostmanMessageTypes.UNRESERVE_REPLY
    ] = PostmanMessageTypes.UNRESERVE_REPLY


class UnreservePubDenied(JSONMessage):
    type: Literal[
        PostmanMessageTypes.UNRESERVE_DENIED
    ] = PostmanMessageTypes.UNRESERVE_DENIED
    error: str


class ReserveSubUpdate(JSONMessage, Reservation):
    type: Literal[
        PostmanSubMessageTypes.RESERVE_UPDATE
    ] = PostmanSubMessageTypes.RESERVE_UPDATE


class AssignList(JSONMessage):
    type: Literal[
        PostmanMessageTypes.LIST_ASSIGNATION
    ] = PostmanMessageTypes.LIST_ASSIGNATION
    exclude: Optional[List[AssignationStatus]]


class AssingListReply(JSONMessage):
    type: Literal[
        PostmanMessageTypes.LIST_ASSIGNATION_REPLY
    ] = PostmanMessageTypes.LIST_ASSIGNATION_REPLY
    assignations: List[Assignation]


class AssignListDenied(JSONMessage):
    type: Literal[
        PostmanMessageTypes.LIST_ASSIGNATION_DENIED
    ] = PostmanMessageTypes.LIST_ASSIGNATION_DENIED
    error: str


class AssignPub(JSONMessage):
    type: Literal[PostmanMessageTypes.ASSIGN] = PostmanMessageTypes.ASSIGN
    reservation: str
    args: List[Any]
    kwargs: Dict[str, Any]
    persist: bool = True
    log: bool = True


class AssignPubReply(JSONMessage, Assignation):
    type: Literal[PostmanMessageTypes.ASSIGN_REPLY] = PostmanMessageTypes.ASSIGN_REPLY
    assignation: str
    status: AssignationStatus


class AssignPubDenied(JSONMessage):
    type: Literal[PostmanMessageTypes.ASSIGN_DENIED] = PostmanMessageTypes.ASSIGN_DENIED
    error: str


class UnassignPub(JSONMessage):
    type: Literal[PostmanMessageTypes.UNASSIGN] = PostmanMessageTypes.UNASSIGN
    assignation: str


class UnassignPubReply(JSONMessage, Unassignation):
    type: Literal[
        PostmanMessageTypes.UNASSIGN_REPLY
    ] = PostmanMessageTypes.UNASSIGN_REPLY


class UnassignPubDenied(JSONMessage):
    type: Literal[
        PostmanMessageTypes.UNASSIGN_DENIED
    ] = PostmanMessageTypes.UNASSIGN_DENIED
    error: str


class AssignSubUpdate(JSONMessage, Assignation):
    type: Literal[
        PostmanSubMessageTypes.ASSIGN_UPDATE
    ] = PostmanSubMessageTypes.ASSIGN_UPDATE



from arkitekt.actors.base import Actor
from arkitekt.agents.transport.base import AgentTransport
from arkitekt.messages import Provision


class ActorBuilder:
    def build_actor(self, provision: Provision, transport: AgentTransport) -> Actor:
        raise NotImplementedError(
            "If you decide to use the mixin you need to implement this method"
        )



import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Awaitable, Callable, Optional

from arkitekt.actors.base import Actor
from arkitekt.actors.helper import AsyncAssignationHelper, ThreadedAssignationHelper
from arkitekt.actors.vars import current_assignation_helper
from arkitekt.api.schema import AssignationStatus
from arkitekt.messages import Assignation, Provision
from arkitekt.structures.serialization.actor import expand_inputs, shrink_outputs
from koil.helpers import iterate_spawned, run_spawned
from pydantic import Field

logger = logging.getLogger(__name__)


class FunctionalActor(Actor):
    assign: Callable[..., Any]
    provide: Optional[Callable[[Provision], Awaitable[Any]]]
    unprovide: Optional[Callable[[], Awaitable[Any]]]


class FunctionalFuncActor(FunctionalActor):
    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def on_assign(self, assignation: Assignation):
        try:
            args, kwargs = (
                await expand_inputs(
                    self.template.node,
                    assignation.args,
                    assignation.kwargs,
                    structure_registry=self.structure_registry,
                )
                if self.expand_inputs
                else (assignation.args, assignation.kwargs)
            )

            await self.transport.change_assignation(
                assignation.assignation,
                status=AssignationStatus.ASSIGNED,
            )

            current_assignation_helper.set(
                AsyncAssignationHelper(
                    actor=self, assignation=assignation, provision=self.provision
                )
            )

            returns = await self.assign(*args, **kwargs)

            current_assignation_helper.set(None)

            returns = (
                await shrink_outputs(
                    self.template.node,
                    returns,
                    structure_registry=self.structure_registry,
                )
                if self.shrink_outputs
                else returns
            )

            await self.transport.change_assignation(
                assignation.assignation,
                status=AssignationStatus.RETURNED,
                returns=returns,
            )

        except asyncio.CancelledError as e:

            await self.transport.change_assignation(
                assignation.assignation, status=AssignationStatus.CANCELLED
            )

        except Exception as e:
            logger.exception(e)
            await self.transport.change_assignation(
                assignation.assignation,
                status=AssignationStatus.CRITICAL,
                message=repr(e),
            )


class FunctionalGenActor(FunctionalActor):
    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def on_assign(self, message: Assignation):
        try:
            args, kwargs = (
                await expand_inputs(
                    self.template.node,
                    message.args,
                    message.kwargs,
                    structure_registry=self.structure_registry,
                )
                if self.expand_inputs
                else (message.args, message.kwargs)
            )

            current_assignation_helper.set(
                AsyncAssignationHelper(
                    actor=self, assignation=message, provision=self.provision
                )
            )

            await self.transport.change_assignation(
                message.assignation,
                status=AssignationStatus.ASSIGNED,
            )

            async for returns in self.assign(*args, **kwargs):

                returns = (
                    await shrink_outputs(
                        self.template.node,
                        returns,
                        structure_registry=self.structure_registry,
                    )
                    if self.shrink_outputs
                    else returns
                )

                await self.transport.change_assignation(
                    message.assignation, status=AssignationStatus.YIELD, returns=returns
                )

            current_assignation_helper.set(None)

            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.DONE
            )

        except asyncio.CancelledError as e:

            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.CANCELLED, message=str(e)
            )

        except Exception as e:
            logger.error("Error in actor", exc_info=True)
            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.CRITICAL, message=str(e)
            )

            raise e


class FunctionalThreadedFuncActor(FunctionalActor):
    threadpool: ThreadPoolExecutor = Field(
        default_factory=lambda: ThreadPoolExecutor(4)
    )

    async def on_assign(self, message: Assignation):

        try:
            logger.info("Assigning Number two")
            args, kwargs = (
                await expand_inputs(
                    self.template.node,
                    message.args,
                    message.kwargs,
                    structure_registry=self.structure_registry,
                )
                if self.expand_inputs
                else (message.args, message.kwargs)
            )

            await self.transport.change_assignation(
                message.assignation,
                status=AssignationStatus.ASSIGNED,
            )

            current_assignation_helper.set(
                ThreadedAssignationHelper(
                    actor=self, assignation=message, provision=self.provision
                )
            )
            returns = await run_spawned(self.assign, *args, **kwargs, pass_context=True)

            current_assignation_helper.set(None)
            shrinked_returns = (
                await shrink_outputs(
                    self.template.node,
                    returns,
                    structure_registry=self.structure_registry,
                )
                if self.expand_inputs
                else (message.args, message.kwargs)
            )

            await self.transport.change_assignation(
                message.assignation,
                status=AssignationStatus.RETURNED,
                returns=shrinked_returns,
            )

        except asyncio.CancelledError as e:
            logger.info("Actor Cancelled")

            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.CANCELLED, message=str(e)
            )

        except Exception as e:
            logger.error("Error in actor", exc_info=True)
            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.CRITICAL, message=str(e)
            )


class FunctionalThreadedGenActor(FunctionalActor):
    threadpool: ThreadPoolExecutor = Field(
        default_factory=lambda: ThreadPoolExecutor(4)
    )

    async def on_assign(self, message: Assignation):
        try:
            args, kwargs = (
                await expand_inputs(
                    self.template.node,
                    message.args,
                    message.kwargs,
                    structure_registry=self.structure_registry,
                )
                if self.expand_inputs
                else (message.args, message.kwargs)
            )

            current_assignation_helper.set(
                ThreadedAssignationHelper(
                    actor=self, assignation=message, provision=self.provision
                )
            )

            await self.transport.change_assignation(
                message.assignation,
                status=AssignationStatus.ASSIGNED,
            )

            async for returns in iterate_spawned(
                self.assign, *args, **kwargs, pass_context=True
            ):

                returns = (
                    await shrink_outputs(
                        self.template.node,
                        returns,
                        structure_registry=self.structure_registry,
                    )
                    if self.shrink_outputs
                    else returns
                )

                await self.transport.change_assignation(
                    message.assignation, status=AssignationStatus.YIELD, returns=returns
                )

            current_assignation_helper.set(None)

            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.DONE
            )

        except asyncio.CancelledError as e:

            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.CANCELLED, message=str(e)
            )

        except Exception as e:
            logger.error("Error in actor", exc_info=True)
            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.CRITICAL, message=str(e)
            )

            raise e



from arkitekt.api.schema import WidgetInput


def SliderWidget(min=0, max=0):
    return WidgetInput(typename="SliderWidget", min=min, max=max)


def SearchWidget(query=""):
    return WidgetInput(typename="SearchWidget", query=query)



import contextvars
from arkitekt.actors.helper import AssignationHelper, ProvisionHelper


current_assignation_helper = contextvars.ContextVar("current_assignation_helper")
current_provision_helper = contextvars.ContextVar("current_provision_helper")


def get_current_assignation_helper() -> AssignationHelper:
    return current_assignation_helper.get()


def get_current_provision_helper() -> ProvisionHelper:
    return current_provision_helper.get()



from typing import Any, Awaitable, Callable, Dict, List
from pydantic import Field
from rich import get_console
from arkitekt.api.schema import TemplateFragment, WidgetInput
from arkitekt.fakts.agent import FaktsAgent
from arkitekt.fakts.postman import FaktsPostman
from arkitekt.fakts.rath import FaktsArkitektRath
from arkitekt.rath import ArkitektRath
from arkitekt.messages import Provision
from arkitekt.structures.default import get_default_structure_registry
from arkitekt.structures.registry import (
    StructureRegistry,
)

from arkitekt.definition.registry import (
    DefinitionRegistry,
    get_current_definition_registry,
)
from arkitekt.agents.base import BaseAgent
from arkitekt.postmans.base import BasePostman
from koil import unkoil
from koil.composition import Composition
from koil.decorators import koilable


@koilable(fieldname="koil", add_connectors=True)
class Arkitekt(Composition):
    rath: ArkitektRath = Field(default_factory=FaktsArkitektRath)
    structure_registry: StructureRegistry = Field(
        default_factory=get_default_structure_registry
    )
    definition_registry: DefinitionRegistry = Field(
        default_factory=get_current_definition_registry
    )
    agent: BaseAgent = Field(default_factory=FaktsAgent)
    postman: BasePostman = Field(default_factory=FaktsPostman)

    def register(
        self,
        builder=None,
        widgets: Dict[str, WidgetInput] = {},
        interfaces: List[str] = [],
        on_provide: Callable[[Provision, TemplateFragment], Awaitable[Any]] = None,
        on_unprovide: Callable[[], Awaitable[Any]] = None,
        structure_registry: StructureRegistry = None,
        **actorparams,
    ) -> None:
        """
        Register a new function
        """
        structure_registry = structure_registry or self.structure_registry

        def real_decorator(function):
            # Simple bypass for now
            def wrapped_function(*args, **kwargs):
                return function(*args, **kwargs)

            self.definition_registry.register(
                function,
                builder=builder,
                widgets=widgets,
                interfaces=interfaces,
                structure_registry=structure_registry,
                on_provide=on_provide,
                on_unprovide=on_unprovide,
                **actorparams,
            )

        return real_decorator

    def run(self, *args, **kwargs) -> None:
        """
        Run the application.
        """
        return unkoil(self.arun, *args, **kwargs)

    async def arun(self) -> None:
        """
        Run the application.
        """
        assert self.agent.transport.connected, "Transport is not connected"
        with get_console().status("Running"):
            await self.agent.aprovide()

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True
        extra = "forbid"



import asyncio
import json

import websockets
from arkitekt.postmans.transport.base import PostmanTransport
from typing import Any, Awaitable, Callable, Dict, List, Optional
from arkitekt.messages import Assignation, Reservation
from arkitekt.postmans.transport.errors import (
    AssignDeniedError,
    AssignationListDeniedError,
    PostmanTransportException,
    ReservationListDeniedError,
    ReserveDeniedError,
    UnassignDeniedError,
    UnreserveDeniedError,
)
from pydantic import Field
from .protocols.postman_json import *
import logging
from websockets.exceptions import ConnectionClosedError, InvalidStatusCode

logger = logging.getLogger(__name__)


async def token_loader():
    raise NotImplementedError(
        "Websocket transport does need a defined token_loader on Connection"
    )


class CorrectableConnectionFail(PostmanTransportException):
    pass


class DefiniteConnectionFail(PostmanTransportException):
    pass


class WebsocketPostmanTransport(PostmanTransport):
    ws_url: str
    token_loader: Callable[[], Awaitable[str]] = Field(exclude=True)
    max_retries = 5
    time_between_retries = 5
    allow_reconnect = True

    _futures: Dict[str, asyncio.Future] = {}
    _connected = False
    _healthy = False
    _send_queue: Optional[asyncio.Queue] = None
    _connection_task: Optional[asyncio.Task] = None

    async def __aenter__(self):
        assert (
            self._abroadcast is not None
        ), "Broadcast must be defined (either overwrite abroadcast or pass this in constructor of transport)"

        assert self.instance_id, "Needs an instance id"
        self._send_queue = asyncio.Queue()
        self._connection_task = asyncio.create_task(self.websocket_loop())
        self._connected = True

    async def websocket_loop(self, retry=0, reload_token=False):
        send_task = None
        receive_task = None
        try:
            try:
                token = await self.token_loader(force_refresh=reload_token)

                async with websockets.connect(
                    f"{self.ws_url}?token={token}&instance_id={self.instance_id}"
                ) as client:

                    logger.info("Postman on Websockets connected")

                    send_task = asyncio.create_task(self.sending(client))
                    receive_task = asyncio.create_task(self.receiving(client))

                    self._healthy = True
                    done, pending = await asyncio.wait(
                        [send_task, receive_task],
                        return_when=asyncio.FIRST_EXCEPTION,
                    )
                    self._healthy = True

                    for task in pending:
                        task.cancel()

                    for task in done:
                        raise task.exception()

            except InvalidStatusCode as e:
                logger.warning(
                    "Websocket Connect was denied. Trying to reload token",
                    exc_info=True,
                )
                reload_token = True
                raise CorrectableConnectionFail from e

            except ConnectionClosedError as e:
                logger.warning("Websocket was closed", exc_info=True)
                raise CorrectableConnectionFail from e

            except Exception as e:
                logger.warning("Websocket excepted. Trying to recover", exc_info=True)
                raise CorrectableConnectionFail from e

        except CorrectableConnectionFail as e:
            logger.info(f"Trying to Recover from Exception {e}")
            if retry > self.max_retries or not self.allow_reconnect:
                raise DefiniteConnectionFail("Exceeded Number of Retries")

            await asyncio.sleep(self.time_between_retries)
            logger.info(f"Retrying to connect")
            await self.websocket_loop(retry=retry + 1, reload_token=reload_token)

        except DefiniteConnectionFail as e:
            logger.error("Websocket excepted closed definetely", exc_info=True)
            self.connection_dead = False
            raise e

        except asyncio.CancelledError as e:
            logger.info("Websocket got cancelled. Trying to shutdown graceully")
            if send_task and receive_task:
                send_task.cancel()
                receive_task.cancel()

            cancellation = await asyncio.gather(
                send_task, receive_task, return_exceptions=True
            )
            raise e

    async def sending(self, client):
        try:
            while True:
                message = await self._send_queue.get()
                await client.send(message)
                self._send_queue.task_done()
        except asyncio.CancelledError as e:
            logger.info("Sending Task sucessfully Cancelled")

    async def receiving(self, client):
        try:
            async for message in client:
                await self.receive(message)
        except asyncio.CancelledError as e:
            logger.info("Receiving Task sucessfully Cancelled")

    async def receive(self, message):
        json_dict = json.loads(message)
        if "type" in json_dict:
            type = json_dict["type"]
            id = json_dict["id"]
            logger.debug(str(json_dict))

            # State Layer
            if type == PostmanSubMessageTypes.ASSIGN_UPDATE:
                await self._abroadcast(AssignSubUpdate(**json_dict))

            if type == PostmanSubMessageTypes.RESERVE_UPDATE:
                await self._abroadcast(ReserveSubUpdate(**json_dict))

            if type == PostmanMessageTypes.LIST_ASSIGNATION_REPLY:
                self._futures[id].set_result(AssingListReply(**json_dict))
            if type == PostmanMessageTypes.LIST_ASSIGNATION_DENIED:
                self._futures[id].set_exception(
                    AssignationListDeniedError(json_dict["error"])
                )

            if type == PostmanMessageTypes.LIST_RESERVATION_REPLY:
                self._futures[id].set_result(ReserveListReply(**json_dict))
            if type == PostmanMessageTypes.LIST_RESERVATION_DENIED:
                self._futures[id].set_exception(
                    ReservationListDeniedError(json_dict["error"])
                )

            if type == PostmanMessageTypes.ASSIGN_REPLY:
                self._futures[id].set_result(AssignPubReply(**json_dict))
            if type == PostmanMessageTypes.ASSIGN_DENIED:
                self._futures[id].set_exception(AssignDeniedError(json_dict["error"]))

            if type == PostmanMessageTypes.UNASSIGN_REPLY:
                self._futures[id].set_result(UnassignPubReply(**json_dict))
            if type == PostmanMessageTypes.UNASSIGN_DENIED:
                self._futures[id].set_exception(UnassignDeniedError(json_dict["error"]))

            if type == PostmanMessageTypes.RESERVE_REPLY:
                self._futures[id].set_result(ReservePubReply(**json_dict))
            if type == PostmanMessageTypes.RESERVE_DENIED:
                self._futures[id].set_exception(ReserveDeniedError(json_dict["error"]))

            if type == PostmanMessageTypes.UNRESERVE_REPLY:
                self._futures[id].set_result(UnreservePubReply(**json_dict))
            if type == PostmanMessageTypes.UNRESERVE_DENIED:
                self._futures[id].set_exception(
                    UnreserveDeniedError(json_dict["error"])
                )

        else:
            logger.error(f"Error {json_dict}")

    async def awaitaction(self, action: JSONMessage):
        assert self._connected, "Websocket not connected"
        if action.id in self._futures:
            raise ValueError("Action already has a future")

        future = asyncio.Future()
        self._futures[action.id] = future
        await self._send_queue.put(action.json())
        return await future

    async def alist_reservations(
        self, exclude: Optional[ReservationStatus] = None
    ) -> List[Reservation]:
        prov_list_reply: ReserveListReply = await self.awaitaction(
            ReserveList(exclude=exclude)
        )
        return prov_list_reply.reservations

    async def aassign(
        self,
        reservation: str,
        args: List[Any],
        kwargs: Dict[str, Any],
        persist=True,
        log=False,
    ) -> Assignation:
        ass_list_reply: AssignPubReply = await self.awaitaction(
            AssignPub(
                reservation=reservation,
                args=args,
                kwargs=kwargs,
                persist=persist,
                log=log,
            )
        )
        return ass_list_reply

    async def areserve(
        self, node: str, params: ReserveParamsInput = None
    ) -> Reservation:
        action = ReservePub(node=node, params=params)
        resrep: ReservePubReply = await self.awaitaction(action)
        return resrep

    async def aunreserve(self, reservation: str) -> Reservation:
        action = UnreservePub(reservation=reservation)
        unres: UnreservePubReply = await self.awaitaction(action)
        return unres

    async def aunassign(self, assignation: str) -> Reservation:
        action = UnassignPub(assignation=assignation)
        unass: UnassignPubReply = await self.awaitaction(action)
        return unass

    async def alist_assignations(
        self, exclude: Optional[AssignationStatus] = None
    ) -> List[Assignation]:
        action = AssignList(exclude=exclude)
        ass_list_reply: AssingListReply = await self.awaitaction(action)
        return ass_list_reply.assignations

    async def __aexit__(self, *args, **kwargs):
        self._connection_task.cancel()

        try:
            await self._connection_task
        except asyncio.CancelledError:
            pass

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True



from typing import Callable
from arkitekt.actors.base import Actor
from arkitekt.agents.transport.base import AgentTransport

from arkitekt.messages import Provision


ActorBuilder = Callable[[Provision, AgentTransport], Actor]



from typing import Any
from qtpy import QtCore
from arkitekt.agents.transport.base import AgentTransport
from arkitekt.messages import Provision
from koil.qt import QtCoro
from arkitekt.actors.functional import FunctionalFuncActor
from arkitekt.actors.builder import ActorBuilder


class QtInLoopBuilder(QtCore.QObject, ActorBuilder):
    """A function that takes a provision and an agent and returns an actor.

    The actor produces by this builder will be running in the same thread as the
    koil instance (aka, the thread that called the builder).

    Args:
        QtCore (_type_): _description_
    """

    def __init__(self, assign=None, *args, parent=None, **actor_kwargs) -> None:
        super().__init__(*args, parent=parent)
        self.coro = QtCoro(
            lambda f, *args, **kwargs: assign(*args, **kwargs), autoresolve=True
        )
        self.provisions = {}
        self.actor_kwargs = actor_kwargs

    async def on_assign(self, *args, **kwargs) -> None:
        return await self.coro.acall(*args)

    def build_actor(self, provision: Provision, transport: AgentTransport) -> Any:
        try:
            ac = FunctionalFuncActor(
                provision=provision,
                transport=transport,
                assign=self.on_assign,
                **self.actor_kwargs
            )
            return ac
        except Exception as e:
            raise e



from arkitekt.api.schema import LogLevelInput
from arkitekt.actors.vars import get_current_assignation_helper


async def alog(message: str, level: LogLevelInput = LogLevelInput.DEBUG) -> None:
    await get_current_assignation_helper().alog(message, level)



from arkitekt.actors.base import Actor
from arkitekt.actors.functional import (
    FunctionalFuncActor,
    FunctionalGenActor,
    FunctionalThreadedFuncActor,
    FunctionalThreadedGenActor,
)

import inspect
from arkitekt.structures.registry import StructureRegistry
from typing import Callable, Optional
import inspect
from .types import ActorBuilder


class ConversionError(Exception):
    pass


class NonConvertableType(ConversionError):
    pass


def isactor(type):
    try:
        if issubclass(type, Actor):
            return True
        else:
            return False
    except Exception as e:
        return False


async def async_none_provide(prov, template):
    return None


async def async_none_unprovide():
    return None


def actify(
    function_or_actor,
    builder: Optional[ActorBuilder] = None,
    bypass_shrink=False,
    bypass_expand=False,
    on_provide=None,
    on_unprovide=None,
    actor_name=None,
    structure_registry: StructureRegistry = None,
    **params,
) -> Callable[[], Actor]:

    if isactor(function_or_actor):
        return function_or_actor

    actor_name = (
        actor_name or f"GeneratedActor{function_or_actor.__name__.capitalize()}"
    )

    is_coroutine = inspect.iscoroutinefunction(function_or_actor)
    is_asyncgen = inspect.isasyncgenfunction(function_or_actor)
    is_method = inspect.ismethod(function_or_actor)

    is_generatorfunction = inspect.isgeneratorfunction(function_or_actor)
    is_function = inspect.isfunction(function_or_actor)

    actor_attributes = {
        "assign": function_or_actor,
        "expand_inputs": not bypass_expand,
        "shrink_outputs": not bypass_shrink,
        "provide": on_provide if on_provide else async_none_provide,
        "provide": on_unprovide if on_unprovide else async_none_unprovide,
        "structure_registry": structure_registry,
    }

    if builder:
        assert hasattr(builder, "build_actor"), "Builder must have a build_actor method"
        instance = builder(**actor_attributes)
        return lambda provision, transport: instance.build_actor(provision, transport)

    if is_coroutine:
        return lambda provision, transport: FunctionalFuncActor(
            provision=provision, transport=transport, **actor_attributes
        )
    elif is_asyncgen:
        return lambda provision, transport: FunctionalGenActor(
            provision=provision, transport=transport, **actor_attributes
        )
    elif is_generatorfunction:
        return lambda provision, transport: FunctionalThreadedGenActor(
            provision=provision, transport=transport, **actor_attributes
        )
    elif is_function or is_method:
        return lambda provision, transport: FunctionalThreadedFuncActor(
            provision=provision, transport=transport, **actor_attributes
        )
    else:
        raise NotImplementedError("No way of converting this to a function")



import contextvars
from arkitekt.api.schema import DefinitionInput
from arkitekt.actors.actify import actify
from arkitekt.definition.define import prepare_definition
from arkitekt.structures.registry import (
    StructureRegistry,
    get_current_structure_registry,
)
from arkitekt.api.schema import WidgetInput
from typing import Dict, List, Callable, Optional, Tuple
from pydantic import Field
from koil.composition import KoiledModel


current_definition_registry = contextvars.ContextVar(
    "current_definition_registry", default=None
)
GLOBAL_DEFINITION_REGISTRY = None


def get_default_definition_registry():
    global GLOBAL_DEFINITION_REGISTRY
    if GLOBAL_DEFINITION_REGISTRY is None:
        GLOBAL_DEFINITION_REGISTRY = DefinitionRegistry()
    return GLOBAL_DEFINITION_REGISTRY


def get_current_definition_registry(allow_global=True):
    return current_definition_registry.get(get_default_definition_registry())


QString = str


class DefinitionRegistry(KoiledModel):
    structure_registry: Optional[StructureRegistry] = None
    defined_nodes: List[Tuple[DefinitionInput, Callable]] = Field(
        default_factory=list, exclude=True
    )
    templated_nodes: List[Tuple[QString, Callable]] = Field(
        default_factory=list, exclude=True
    )
    copy_from_default: bool = False

    _token: contextvars.Token = None

    def __post_init__(self):
        if self.copy_from_default:
            default = get_default_definition_registry()
            self.defined_nodes = default.defined_nodes + self.defined_nodes
            self.templated_nodes = default.templated_nodes + self.templated_nodes

    def has_definitions(self):
        return len(self.definedNodes) > 0 or len(self.templatedNodes) > 0

    def reset(self):
        self.definedNodes = []  # dict are queryparams for the node
        self.templatedNodes = []

    def register_actor_with_defintion(
        self, actorBuilder: Callable, definition: DefinitionInput, **params  # New Node
    ):
        self.defined_nodes.append((definition, actorBuilder, params))

    def register_actor_with_template(
        self, actorBuilder: Callable, q_string: QString, **params
    ):  # Query Path
        self.templated_nodes.append((q_string, actorBuilder, params))

    def register(
        self,
        function,
        builder=None,
        widgets: Dict[str, WidgetInput] = {},
        interfaces: List[str] = [],
        on_provide=None,
        on_unprovide=None,
        structure_registry: StructureRegistry = None,
        **actorparams,
    ):

        structure_registry = (
            structure_registry
            or self.structure_registry
            or get_current_structure_registry()
        )

        actorBuilder = actify(
            function,
            builder=builder,
            on_provide=on_provide,
            on_unprovide=on_unprovide,
            structure_registry=structure_registry,
            **actorparams,
        )

        definition = prepare_definition(
            function=function,
            widgets=widgets,
            interfaces=interfaces,
            structure_registry=structure_registry,
        )

        self.register_actor_with_defintion(actorBuilder, definition, **actorparams)

    def template(
        self,
        function,
        qstring: QString,
        on_provide=None,
        on_unprovide=None,
        structure_registry: StructureRegistry = None,
        **actorparams,
    ):
        structure_registry = (
            structure_registry
            or self.structure_registry
            or get_current_structure_registry()
        )

        actorBuilder = actify(
            function,
            on_provide=on_provide,
            on_unprovide=on_unprovide,
            structure_registry=structure_registry ** actorparams,
        )

        self.register_actor_with_template(actorBuilder, qstring, **actorparams)

    async def __aenter__(self):
        self.structure_registry = (
            self.structure_registry or get_current_structure_registry()
        )
        current_definition_registry.set(self)
        return self

    async def __exit__(self, *args, **kwargs):
        current_definition_registry.set(None)


def register(
    widgets: Dict[str, WidgetInput] = {},
    interfaces: List[str] = [],
    on_provide=None,
    on_unprovide=None,
    definition_registry: DefinitionRegistry = None,
    structure_registry: StructureRegistry = None,
    **params,
):
    """Take a function and register it as a node.

    This function is used to register a node. Use it as a decorator. You can specify
    specific widgets for every paramer in a dictionary {argument_key: widget}. By default
    this function will use the default defintion registry to store the nodes inputdata.
    This definition registry will then be used by an agent to create, and provide the node.

    If your function has specific inputs that need custom rules for expansion and shrinking
     , you can pass a structure registry to the function. This registry will then be used.

    This decorator is non intrusive. You can still call this function as a normal function from
    your code

    Args:
        widgets (Dict[str, WidgetInput], optional): _description_. Defaults to {}.
        interfaces (List[str], optional): _description_. Defaults to [].
        on_provide (_type_, optional): _description_. Defaults to None.
        on_unprovide (_type_, optional): _description_. Defaults to None.
        definition_registry (DefinitionRegistry, optional): _description_. Defaults to None.
        structure_registry (StructureRegistry, optional): _description_. Defaults to None.

    Returns:
        Callable: A wrapped function that just returns the original function.
    """
    definition_registry = definition_registry or get_current_definition_registry()
    structure_registry = structure_registry or get_current_structure_registry()

    def real_decorator(function):
        # Simple bypass for now
        def wrapped_function(*args, **kwargs):
            return function(*args, **kwargs)

        definition_registry.register(
            function,
            widgets=widgets,
            interfaces=interfaces,
            structure_registry=structure_registry,
            on_provide=on_provide,
            on_unprovide=on_unprovide,
            **params,
        )

    return real_decorator


def template(
    q_string: QString,
    on_provide=None,
    on_unprovide=None,
    structure_registry: StructureRegistry = None,
    definition_registry: DefinitionRegistry = None,
    **params,
):

    definition_registry = definition_registry or get_current_definition_registry()
    structure_registry = structure_registry or get_current_structure_registry()

    def real_decorator(function):
        # Simple bypass for now
        def wrapped_function(*args, **kwargs):
            return function(*args, **kwargs)

        definition_registry.template(
            function,
            q_string,
            on_provide=on_provide,
            on_unprovide=on_unprovide,
            structure_registry=structure_registry,
            **params,
        )

        # We are registering this as a template

        return wrapped_function

    return real_decorator



from typing import Awaitable, Callable, List, Optional, Dict, Any, Union

from arkitekt.messages import Assignation, Reservation, Unassignation, Unreservation
from arkitekt.api.schema import AssignationStatus, ReservationStatus, ReserveParamsInput
from koil.composition import KoiledModel


class PostmanTransport(KoiledModel):
    instance_id: Optional[str]

    _abroadcast: Optional[Callable[[Union[Assignation, Reservation]], Awaitable[None]]]

    async def aassign(
        self,
        reservation: str,
        args: List[Any],
        kwargs: Dict[str, Any],
        persist=True,
        log=False,
    ) -> Assignation:
        raise NotImplementedError()

    async def aunassign(self, assignation: str) -> Unassignation:
        raise NotImplementedError()

    async def areserve(
        self,
        node: str,
        params: ReserveParamsInput = None,
    ) -> Reservation:
        raise NotImplementedError()

    async def aunreserve(
        self,
        reservation: str,
    ) -> Unreservation:
        raise NotImplementedError()

    async def alist_assignations(
        self, exclude: Optional[AssignationStatus] = None
    ) -> List[Assignation]:
        raise NotImplementedError()

    async def alist_reservations(
        self, exclude: Optional[ReservationStatus] = None
    ) -> List[Reservation]:
        raise NotImplementedError()

    class Config:
        underscore_attrs_are_private = True
